# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading core plasma profiles from the core_profiles IDS."""

from collections.abc import Callable

import numpy as np
from raysect.core.math import Vector3D, translate
from raysect.core.math.function.float import Function2D, Function3D, Interpolator1DArray
from raysect.core.math.function.vector3d import Constant3D as ConstantVector3D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.primitive import Cylinder, Subtract
from scipy.constants import atomic_mass, electron_mass

from cherab.core import Maxwellian, Plasma, Species
from cherab.core.math import VectorAxisymmetricMapper
from cherab.core.utility import RecursiveDict
from cherab.tools.equilibrium import EFITEquilibrium
from imas import DBEntry

from ..ids.common import get_ids_time_slice
from ..ids.core_profiles import load_core_grid, load_core_species
from .equilibrium import load_equilibrium, load_magnetic_field
from .utility import warn_unsupported_species

__all__ = ["load_core_plasma"]


def load_core_plasma(
    *args,
    time: float = 0,
    occurrence: int = 0,
    equilibrium: EFITEquilibrium | None = None,
    b_field: VectorFunction2D | None = None,
    psi_interpolator: Callable[[float], float] | None = None,
    time_threshold: float = np.inf,
    parent=None,
    **kwargs,
) -> Plasma:
    """Load core profiles and Create a `~cherab.core.plasma.node.Plasma` object.

    Prefer 'density_thermal' over 'density' profile.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time : float, optional
        Time moment, by default 0.
    occurrence : int, optional
        Instance index of the 'core_profiles' IDS, by default 0.
    equilibrium : EFITEquilibrium, optional
        Alternative `~cherab.tools.equilibrium.efit.EFITEquilibrium` object used to map core
        profiles, by default None. equilibrium is read from the same IMAS query as the core profiles.
        This parameter is ignored if core plasma is not available.
    b_field : VectorFunction2D, optional
        An alternative 2D interpolator of the magnetic field vector (Br, Btor, Bz).
        Default is None. The magnetic field will be loaded from the 'equilibrium' IDS.
    psi_interpolator : Callable[[float], float], optional
        Alternative `psi_norm(rho_tor_norm)` interpolator.
        Used only if 'psi' is missing in the core grid, by default None.
        Obtained from the 'equilibrium' IDS.
    time_threshold : float, optional
        Sets the maximum allowable difference between the specified time and the nearest
        available time, by default `numpy.inf`.
    parent : _NodeBase, optional
        Parent node in the Raysect scene-graph, by default None.
        Normally, `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `~cherab.core.plasma.node.Plasma`
        Plasma object with core profiles.
    """

    with DBEntry(*args, **kwargs) as entry:
        core_profiles_ids = get_ids_time_slice(
            entry, "core_profiles", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    if not len(core_profiles_ids.profiles_1d):
        raise RuntimeError("The profiles_1d AOS in core_profiles IDS is empty.")

    if equilibrium is None:
        equilibrium, psi_interp = load_equilibrium(
            *args, time=time, occurrence=occurrence, with_psi_interpolator=True, **kwargs
        )
        psi_interpolator = psi_interpolator or psi_interp

    if not isinstance(equilibrium, EFITEquilibrium):
        raise ValueError("Argiment equilibrium must be a EFITEquilibrium instance.")

    if b_field is None:
        try:
            b_field = load_magnetic_field(*args, time=time, occurrence=occurrence, **kwargs)
        except RuntimeError:
            b_field = equilibrium.b_field

    core_grid = load_core_grid(core_profiles_ids.profiles_1d[0].grid)

    composition = load_core_species(core_profiles_ids.profiles_1d[0])

    psi_norm = get_psi_norm(
        core_grid["psi"],
        equilibrium.psi_axis,
        equilibrium.psi_lcfs,
        core_grid["rho_tor_norm"],
        psi_interpolator,
    )

    name = f"IMAS core plasma: time {core_profiles_ids.time[0]}, uri {entry.uri}."
    plasma = Plasma(parent=parent, name=name)

    # Create plasma geometry
    radius_inner, radius_outer = equilibrium.r_range
    zmin, zmax = equilibrium.z_range
    height = zmax - zmin
    plasma.geometry = Subtract(Cylinder(radius_outer, height), Cylinder(radius_inner, height))
    plasma.geometry_transform = translate(0, 0, zmin)

    plasma.b_field = VectorAxisymmetricMapper(b_field)

    # Add electron species
    electrons = get_core_interpolators(
        psi_norm, composition["electron"], equilibrium, return3d=True
    )
    if electrons["density_thermal"] is not None:
        electrons["density"] = electrons["density_thermal"]

    if electrons["density"] is None:
        print("Unable to create Core Plasma: electron density is not available.")
    if electrons["temperature"] is None:
        print("Unable to create Core Plasma: electron temperature is not available.")

    zero_velocity = ConstantVector3D(Vector3D(0, 0, 0))

    plasma.electron_distribution = Maxwellian(
        electrons["density"], electrons["temperature"], zero_velocity, electron_mass
    )

    warn_unsupported_species(composition, "molecule")
    warn_unsupported_species(composition, "molecular_bundle")
    warn_unsupported_species(composition, "ion_bundle")

    # Add ion and neutral species
    for species_id, profiles in composition["ion"].items():
        d = {first: second for first, second in species_id}
        species_type = d["element"]
        charge = int(round(d["z"]))

        sp_key = (species_type, charge)
        if sp_key in plasma.composition:
            print(
                f"Warning! Skipping {d['name']} species. "
                + f"Species with the same (element, charge): {sp_key} is already added."
            )
            continue

        interp = get_core_interpolators(psi_norm, profiles, equilibrium, return3d=True)
        if interp["density_thermal"] is not None:
            interp["density"] = interp["density_thermal"]

        if interp["density"] is None:
            print(f"Warning! Skipping {d['name']} species: density is not available.")
        if interp["temperature"] is None:
            print(f"Warning! Skipping {d['name']} species: temperature is not available.")

        distribution = Maxwellian(
            interp["density"],
            interp["temperature"],
            zero_velocity,
            species_type.atomic_weight * atomic_mass,
        )

        plasma.composition.add(Species(species_type, charge, distribution))

    return plasma


def get_core_interpolators(
    psi_norm: np.ndarray,
    profiles: dict,
    equilibrium: EFITEquilibrium,
    return3d: bool = False,
) -> dict[str, Function3D | Function2D | None]:
    """Create interpolators for the core profiles.

    Parameters
    ----------
    psi_norm : ndarray
        Normalized poloidal flux values.
    profiles : dict
       Dictionary with core plasma profiles.
    equilibrium : EFITEquilibrium
        `EFITEquilibrium` object used to map core profiles.
    return3d : bool, optional
        If True, return the 3D interpolators assuming rotational symmetry, by default False.

    Returns
    -------
    dict[str, Function3D | Function2D | None]
       Dictionary with core interpolators.
    """

    if not isinstance(equilibrium, EFITEquilibrium):
        raise ValueError("Argiment equilibrium must be a EFITEquilibrium instance.")

    psi_norm, index = np.unique(psi_norm, return_index=True)

    interpolators = RecursiveDict()

    for prof_key, profile in profiles.items():
        if profile is not None:
            extrapolation_range = max(0, psi_norm[0], 1.0 - psi_norm[-1])
            func = Interpolator1DArray(
                psi_norm, profile[index], "cubic", "nearest", extrapolation_range
            )
            interpolators[prof_key] = (
                equilibrium.map3d(func) if return3d else equilibrium.map2d(func)
            )
        else:
            interpolators[prof_key] = None

    return interpolators.freeze()


def get_psi_norm(
    psi: np.ndarray | None,
    psi_axis: float,
    psi_lcfs: float,
    rho_tor_norm: np.ndarray | None,
    psi_interpolator: Callable[[float], float] | None,
) -> np.ndarray:
    """Calculate normalized poloidal flux.

    Parameters
    ----------
    psi : ndarray | None
        Poloidal flux values from the core grid.
    psi_axis : float
        Poloidal flux at the magnetic axis.
    psi_lcfs : float
        Poloidal flux at the last closed flux surface.
    rho_tor_norm : ndarray | None
        Normalized toroidal flux values.
    psi_interpolator : Callable[[float], float] | None
        Interpolator function to map rho_tor_norm to psi_norm.
        Used only if 'psi' is None.

    Returns
    -------
    ndarray
        Normalized poloidal flux values.
    """
    if psi is None:
        if psi_interpolator is None:
            raise RuntimeError(
                "Unable to map rho_tor_norm to psi_norm grid: psi_interpolator is not provided."
            )

        if rho_tor_norm is None:
            raise RuntimeError(
                "No rho_tor_norm values are available in the core grid: unable to interpolate to psi_norm."
            )

        return np.array([psi_interpolator(rho) for rho in rho_tor_norm])

    return (-psi / (2 * np.pi) - psi_axis) / (psi_lcfs - psi_axis)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for common functions used to get IDS time slices."""

import warnings

from numpy import inf

from imas import DBEntry
from imas.ids_defs import CLOSEST_INTERP
from imas.ids_toplevel import IDSToplevel

__all__ = ["get_ids_time_slice"]


def get_ids_time_slice(
    entry: DBEntry,
    ids_name: str,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = inf,
) -> IDSToplevel:
    """Get a time slice of the specified IDS from the given IMAS entry.

    .. note::
        If the `~imas.db_entry.DBEntry.get_slice` method is not implemented for the given IMAS entry
        URI, this function will fall back to using the `~imas.db_entry.DBEntry.get` method and
        return the entire IDS.

    Parameters
    ----------
    entry : `~imas.db_entry.DBEntry`
        The IMAS entry. The entry must be opened in read mode.
    ids_name : str
        The name of the IDS.
    time : float, optional
        The time in seconds of the requested time slice, by default is 0.
    occurrence : int, optional
        The occurrence of the IDS, by default is 0.
    time_threshold : float, optional
        The maximum allowed time difference in seconds between the actual time of the nearest time
        slice and the given time, by default is infinity.

    Returns
    -------
    `~imas.ids_toplevel.IDSToplevel`
        The requested IDS time slice.
    """
    if time < 0:
        raise ValueError(f"Argument 'time' must be >=0 ({time} s).")
    if time_threshold < 0:
        raise ValueError(f"Argument 'time_threshold' must be >=0 ({time_threshold} s).")

    try:
        ids = entry.get_slice(
            ids_name,
            time,
            CLOSEST_INTERP,
            occurrence=occurrence,
        )
    except NotImplementedError:
        # Fallback to `get` method to retrieve the entire IDS
        warnings.warn(
            f"The 'get_slice' method is not implemented for the URI '{entry.uri}'. "
            + "Falling back to 'get' method to retrieve the entire IDS.",
            RuntimeWarning,
            stacklevel=2,
        )
        ids = entry.get(ids_name, occurrence=occurrence)

    if not len(ids.time):
        raise RuntimeError(f"The '{ids_name}' IDS is empty.")

    if abs(ids.time[0] - time) > time_threshold:
        raise RuntimeError(
            f"The time difference between the actual time ({ids.time[0]} s) "
            + f"of the nearest '{ids_name}' time slice and the given time ({time} s) "
            + f"exceeds the specified threshold ({time_threshold} s)."
        )

    return ids



import pytest
from imas import DBEntry
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from raysect.core.scenegraph import World
from raysect.primitive import Subtract

from cherab.core import Maxwellian, Plasma
from cherab.imas.ids.common import get_ids_time_slice
from cherab.imas.ids.common.ggd import load_grid
from cherab.imas.ids.edge_profiles import load_edge_species
from cherab.imas.plasma.edge import load_edge_plasma
from cherab.imas.plasma.equilibrium import load_magnetic_field

plt.rcParams["backend"] = "Agg"  # Use non-interactive backend for testing


def test_load_edge_plasma_basic(path_iter_jintrac: str):
    """Test basic loading of edge plasma data from an IMAS file."""
    plasma = load_edge_plasma(path_iter_jintrac, "r")

    # Test that a Plasma object is returned
    assert isinstance(plasma, Plasma)


def test_load_edge_plasma_geometry(path_iter_jintrac: str):
    """Test that the plasma geometry is properly set."""
    plasma = load_edge_plasma(path_iter_jintrac, "r")

    # Check that geometry is set and has the expected type
    assert plasma.geometry is not None
    assert isinstance(plasma.geometry, Subtract)

    # Check that geometry transform is set
    assert plasma.geometry_transform is not None


def test_load_edge_plasma_electron_distribution(path_iter_jintrac: str):
    """Test that electron distribution is properly loaded."""
    plasma = load_edge_plasma(path_iter_jintrac, "r")

    # Check that electron distribution exists
    assert plasma.electron_distribution is not None
    assert isinstance(plasma.electron_distribution, Maxwellian)


def test_load_edge_plasma_with_time(path_iter_jintrac: str):
    """Test loading edge plasma with specific time parameter."""
    # Test with default time (should not raise error)
    plasma1 = load_edge_plasma(path_iter_jintrac, "r", time=0.0)
    assert isinstance(plasma1, Plasma)

    # Test with different time values
    plasma2 = load_edge_plasma(path_iter_jintrac, "r", time=0.1)
    assert isinstance(plasma2, Plasma)


def test_load_edge_plasma_with_grid_subset_id(path_iter_jintrac: str):
    """Test loading edge plasma with different grid subset IDs."""
    # Test with string identifier
    plasma2 = load_edge_plasma(path_iter_jintrac, "r", grid_subset_id="cells")
    assert isinstance(plasma2, Plasma)


def test_load_edge_plasma_with_parent(path_iter_jintrac: str):
    """Test loading edge plasma with a parent scene graph node."""
    world = World()

    plasma = load_edge_plasma(path_iter_jintrac, "r", parent=world)
    assert isinstance(plasma, Plasma)
    assert plasma.parent is world


def test_load_edge_plasma_with_magnetic_field(path_iter_jintrac: str):
    """Test loading edge plasma with external magnetic field."""
    # Load magnetic field separately
    try:
        b_field = load_magnetic_field(path_iter_jintrac, "r")
        plasma = load_edge_plasma(path_iter_jintrac, "r", b_field=b_field)
        assert isinstance(plasma, Plasma)
        assert plasma.b_field is not None
    except RuntimeError:
        # If magnetic field loading fails, test should still pass
        # as this is handled gracefully in the function
        pass


def test_load_edge_plasma_time_threshold(path_iter_jintrac: str):
    """Test loading edge plasma with time threshold parameter."""
    # Test with different time thresholds
    plasma = load_edge_plasma(
        path_iter_jintrac, "r", time_threshold=500
    )  # test dataset has time ~450s only
    assert isinstance(plasma, Plasma)

    with pytest.raises(RuntimeError):
        load_edge_plasma(path_iter_jintrac, "r", time_threshold=0.0)


def test_load_edge_plasma_composition(path_iter_jintrac: str):
    """Test that plasma composition contains expected species."""
    plasma = load_edge_plasma(path_iter_jintrac, "r")

    # Check that we have some ion species (exact species depend on data)
    assert len(plasma.composition) > 0


def test_load_edge_plasma_error_handling(path_iter_jintrac: str):
    """Test error handling for invalid inputs."""
    # Test with non-existent file path
    with pytest.raises((FileNotFoundError, RuntimeError, OSError)):
        load_edge_plasma("non_existent_path", "r")

    # Test with invalid mode
    with pytest.raises((ValueError, RuntimeError, OSError)):
        load_edge_plasma(path_iter_jintrac, "invalid_mode")


def test_plot_edge_plasma_profiles(path_iter_jintrac: str):
    """Test mesh plotting of edge plasma profiles."""
    with DBEntry(path_iter_jintrac, "r") as entry:
        ids = get_ids_time_slice(entry, "edge_profiles", time=0)

    # Load mesh grid
    grid, subsets, subset_id = load_grid(ids.grid_ggd[0], with_subsets=True)
    grid_subset_name = "Cells"
    try:
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)
    except KeyError:
        grid_subset_name = grid_subset_name.lower()
        grid = grid.subset(subsets[grid_subset_name.lower()], name=grid_subset_name)

    # Load edge species composition
    composition = load_edge_species(ids.ggd[0], grid_subset_index=subset_id[grid_subset_name])

    # Plot electron density
    electron_density = composition["electron"]["density"]
    ax = grid.plot_mesh(data=electron_density)

    assert isinstance(ax, Axes)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading magnetic field data from the equilibrium IDS."""

import numpy as np

from imas.ids_defs import EMPTY_INT
from imas.ids_struct_array import IDSStructArray

RECTANGULAR_GRID = 1

__all__ = ["load_magnetic_field_data"]


def load_magnetic_field_data(profiles_2d: IDSStructArray) -> dict:
    """Load 2D profiles of the magnetic field components from the profiles_2d structure of the
    equilibrium IDS.

    Parameters
    ----------
    profiles_2d : IDSStructArray
        The profiles_2d structure of the equilibrium IDS.

    Returns
    -------
    dict[str, ndarray]
        Dictionary with the following keys:

        :r: (N,) ndarray with R coordinates of rectangular grid.
        :z: (M,) ndarray with Z coordinates of rectangular grid.
        :b_field_r: (N, M) ndarray with R component of the magnetic field.
        :b_field_z: (N, M) ndarray with Z component of the magnetic field.
        :b_field_phi: (N, M) ndarray with toroidal component of the magnetic field.
    """

    rectangular_grid = False
    for prof2d in profiles_2d:
        if prof2d.grid_type.index == RECTANGULAR_GRID or prof2d.grid_type.index == EMPTY_INT:
            rectangular_grid = True
            break

    if not rectangular_grid:
        raise RuntimeError(
            "Unable to read magnetic field: "
            + "rectangular grid for 2D profiles is not found and other grid types are not supported."
        )

    b_dict = {}

    b_dict["r"] = np.asarray_chkfinite(prof2d.grid.dim1)
    b_dict["z"] = np.asarray_chkfinite(prof2d.grid.dim2)
    shape = (b_dict["r"].size, b_dict["z"].size)

    b_dict["b_field_r"] = np.asarray_chkfinite(prof2d.b_field_r)
    b_dict["b_field_phi"] = np.asarray_chkfinite(prof2d.b_field_phi)
    b_dict["b_field_z"] = np.asarray_chkfinite(prof2d.b_field_z)

    if (
        b_dict["b_field_r"].shape != shape
        or b_dict["b_field_phi"].shape != shape
        or b_dict["b_field_z"].shape != shape
    ):
        raise RuntimeError(
            "Unable to read magnetic field: "
            + "the shape of the magnetic field components does not match the grid shape."
        )

    return b_dict



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module to offer the plasma object, blending core with edge profiles."""

from collections.abc import Callable

import numpy as np
from raysect.core.math import Vector3D, translate
from raysect.core.math.function.float import Blend2D, Blend3D, Function2D, Function3D
from raysect.core.math.function.vector3d import Blend2D as VectorBlend2D
from raysect.core.math.function.vector3d import Blend3D as VectorBlend3D
from raysect.core.math.function.vector3d import Constant3D as ConstantVector3D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.math.function.vector3d import Function3D as VectorFunction3D
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.primitive import Cylinder, Subtract
from scipy.constants import atomic_mass, electron_mass

from cherab.core import Maxwellian, Plasma, Species
from cherab.core.math import AxisymmetricMapper, VectorAxisymmetricMapper
from cherab.tools.equilibrium import EFITEquilibrium
from imas import DBEntry
from imas.ids_structure import IDSStructure

from ..ids.common import get_ids_time_slice
from ..ids.common.ggd import load_grid
from ..ids.core_profiles import load_core_grid, load_core_species
from ..ids.edge_profiles import load_edge_species
from .core import get_core_interpolators, get_psi_norm, load_core_plasma
from .edge import get_edge_interpolators, load_edge_plasma
from .equilibrium import load_equilibrium, load_magnetic_field
from .utility import get_subset_name_index, warn_unsupported_species

__all__ = ["load_plasma"]


def load_plasma(
    *args,
    time: float = 0,
    occurrence_core: int = 0,
    edge_args: tuple | None = None,
    edge_kwargs: dict | None = None,
    time_edge: float | None = None,
    occurrence_edge: int = 0,
    grid_ggd: IDSStructure | None = None,
    grid_subset_id: int | str = 5,
    equilibrium: EFITEquilibrium | None = None,
    b_field: VectorFunction2D | None = None,
    psi_interpolator: Callable[[float], float] | None = None,
    mask: Function2D | Function3D | None = None,
    time_threshold: float = np.inf,
    parent: _NodeBase | None = None,
    **kwargs,
) -> Plasma:
    """Load core and edge profiles and Create a `~cherab.core.plasma.node.Plasma` object.

    If ``edge_profiles`` IDS is empty, this returns core plasma only.
    If ``core_profiles`` IDS is empty, this returns edge plasma only.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time : float, optional
        Time moment for the core plasma, by default 0.
    occurrence_core : int, optional
        Instance index of the 'core_profiles' IDS, by default 0.
    edge_args : tuple, optional
        Arguments passed to the `~imas.db_entry.DBEntry` constructor for the edge plasma if different
        from the core plasma, by default None: use the same as for the core plasma.
    edge_kwargs : dict, optional
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor for the edge plasma if different
        from the core plasma, by default None: use the same as for the core plasma.
    time_edge : float, optional
        Time moment for the edge plasma if different from the 'time', by default None.
    occurrence_edge : int, optional
        Instance index of the 'edge_profiles' IDS, by default 0.
    grid_ggd : IDSStructure, optional
        Alternative grid_ggd structure with the grid description, by default None.
    grid_subset_id : int | str, optional
        Identifier of the grid subset. Either index or name, by default 5 ("Cells").
    equilibrium : EFITEquilibrium, optional
        Alternative `~cherab.tools.equilibrium.efit.EFITEquilibrium` object used to map core
        profiles, by default None. equilibrium is read from the same IMAS query as the core profiles.
        This parameter is ignored if core plasma is not available.
    b_field : VectorFunction2D, optional
        Alternative 2D interpolator of the magnetic field vector (Br, Btor, Bz).
        Default is None. The magnetic field will be loaded from the 'equilibrium' IDS.
    psi_interpolator : Callable[[float], float], optional
        Alternative `psi_norm(rho_tor_norm)` interpolator.
        Used only if 'psi' is missing in the core grid, by default None.
        Obtained from the 'equilibrium' IDS.
    mask : Function2D | Function3D, optional
        Mask function used for blending: ``(1 - mask) * f_edge + mask * f_core``.
        Default is None. Use `EFITEquilibrium.inside_lcfs` as a mask function.
    time_threshold : float, optional
        Sets the maximum allowable difference between the specified time and the nearest
        available time, by default `numpy.inf`.
    parent : _NodeBase, optional
        Parent node in the Raysect scene-graph, by default None.
        Normally, `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `~cherab.core.plasma.node.Plasma`
        Plasma object with core and edge profiles.
    """

    edge_args = edge_args or args
    edge_kwargs = edge_kwargs or kwargs
    if time_edge is None:
        time_edge = time

    try:
        with DBEntry(*args, **kwargs) as entry_core:
            core_profiles_ids = get_ids_time_slice(
                entry_core,
                "core_profiles",
                time=time,
                occurrence=occurrence_core,
                time_threshold=time_threshold,
            )
    except RuntimeError:
        return load_edge_plasma(
            *edge_args,
            time=time_edge,
            occurrence=occurrence_edge,
            grid_ggd=grid_ggd,
            grid_subset_id=grid_subset_id,
            b_field=b_field,
            time_threshold=time_threshold,
            parent=parent,
            **edge_kwargs,
        )

    try:
        with DBEntry(*edge_args, **edge_kwargs) as entry_edge:
            edge_profiles_ids = get_ids_time_slice(
                entry_edge,
                "edge_profiles",
                time=time_edge,
                occurrence=occurrence_edge,
                time_threshold=time_threshold,
            )
    except RuntimeError:
        return load_core_plasma(
            *args,
            time=0,
            occurrence=0,
            equilibrium=equilibrium,
            b_field=b_field,
            psi_interpolator=psi_interpolator,
            time_threshold=time_threshold,
            parent=parent,
            **kwargs,
        )

    if not len(core_profiles_ids.profiles_1d):
        raise RuntimeError("The profiles_1d AOS in core_profiles IDS is empty.")

    if not len(edge_profiles_ids.grid_ggd) and grid_ggd is None:
        raise RuntimeError(
            "The 'grid_ggd' AOS of the edge_profiles IDS is empty "
            + "and an alternative grid_ggd structure is not provided."
        )

    if not len(edge_profiles_ids.ggd):
        raise RuntimeError("The 'ggd' AOS of the edge_profiles IDS is empty.")

    # Getting equilibrium and magnetic field
    if equilibrium is None:
        equilibrium, psi_interp = load_equilibrium(
            *args, time=time, with_psi_interpolator=True, **kwargs
        )
        psi_interpolator = psi_interpolator or psi_interp

    if b_field is None:
        try:
            b_field = load_magnetic_field(*args, time=time, **kwargs)
        except RuntimeError:
            b_field = equilibrium.b_field

    mask = mask or equilibrium.inside_lcfs

    # Getting core data
    core_grid = load_core_grid(core_profiles_ids.profiles_1d[0].grid)

    composition_core = load_core_species(core_profiles_ids.profiles_1d[0])

    psi_norm = get_psi_norm(
        core_grid["psi"],
        equilibrium.psi_axis,
        equilibrium.psi_lcfs,
        core_grid["rho_tor_norm"],
        psi_interpolator,
    )

    # Getting edge data
    grid_ggd = grid_ggd or edge_profiles_ids.grid_ggd[0]
    grid, subsets, subset_id = load_grid(grid_ggd, with_subsets=True)

    grid_subset_name, grid_subset_index = get_subset_name_index(subset_id, grid_subset_id)

    if np.all(subsets[grid_subset_name] != np.arange(grid.num_cell, dtype=int)):
        # To reduce memory usage, create the sub-grid only if needed.
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)

    composition_edge = load_edge_species(
        edge_profiles_ids.ggd[0], grid_subset_index=grid_subset_index
    )

    # Creating plasma
    tcore = core_profiles_ids.time[0]
    tedge = edge_profiles_ids.time[0]
    name = (
        f"IMAS core + edge plasma: core/edge time {tcore}/{tedge}, "
        f"uri {entry_core.uri} / {entry_edge.uri}."
    )
    plasma = Plasma(parent=parent, name=name)

    # Create plasma geometry
    radius_outer = grid.mesh_extent["rmax"]
    radius_inner = grid.mesh_extent["rmin"]
    height = grid.mesh_extent["zmax"] - grid.mesh_extent["zmin"]
    zmin = grid.mesh_extent["zmin"]
    plasma.geometry = Subtract(Cylinder(radius_outer, height), Cylinder(radius_inner, height))
    plasma.geometry_transform = translate(0, 0, zmin)

    plasma.b_field = VectorAxisymmetricMapper(b_field)

    # Add electron species
    electrons_core = get_core_interpolators(
        psi_norm, composition_core["electron"], equilibrium, return3d=False
    )
    if electrons_core["density_thermal"] is not None:
        electrons_core["density"] = electrons_core["density_thermal"]

    electrons_edge = get_edge_interpolators(
        grid, composition_edge["electron"], b_field, return3d=False
    )

    electrons = blend_core_edge_interpolators(electrons_core, electrons_edge, mask, return3d=True)

    if electrons["density"] is None:
        print("Unable to create Core Plasma: electron density is not available.")
    if electrons["temperature"] is None:
        print("Unable to create Core Plasma: electron temperature is not available.")

    plasma.electron_distribution = Maxwellian(
        electrons["density"], electrons["temperature"], electrons["velocity"], electron_mass
    )

    warn_unsupported_species(composition_core, "molecule")
    warn_unsupported_species(composition_edge, "molecule")
    warn_unsupported_species(composition_core, "molecular_bundle")
    warn_unsupported_species(composition_edge, "molecular_bundle")
    warn_unsupported_species(composition_core, "ion_bundle")
    warn_unsupported_species(composition_edge, "ion_bundle")

    # -----------------------------------
    # === Add ion & neutral species ===
    # -----------------------------------
    # List core species
    core_species = {}
    for species_id, profiles in composition_core["ion"].items():
        d = {first: second for first, second in species_id}
        sp_key = (d["element"], int(round(d["z"])))
        if sp_key in core_species:
            print(
                f"Warning! Skipping {d['name']} core species. "
                + f"Species with the same (element, charge): {sp_key} is already added."
            )
            continue
        if profiles["density_thermal"] is not None:
            profiles["density"] = profiles["density_thermal"]
        core_species[sp_key] = profiles

    # List edge species
    edge_species = {}
    for species_id, profiles in composition_edge["ion"].items():
        d = {first: second for first, second in species_id}
        sp_key = (d["element"], int(round(d["z"])))
        if sp_key in edge_species:
            print(
                f"Warning! Skipping {d['name']} edge species. "
                + f"Species with the same (element, charge): {sp_key} is already added."
            )
            continue
        edge_species[sp_key] = profiles

    species = {}
    for core_key, core_profiles in core_species.items():
        if core_key in edge_species:
            edge_profiles = edge_species[core_key]
            core_interp = get_core_interpolators(
                psi_norm, core_profiles, equilibrium, return3d=False
            )
            edge_interp = get_edge_interpolators(grid, edge_profiles, b_field, return3d=False)
            species[core_key] = blend_core_edge_interpolators(
                core_interp, edge_interp, mask, return3d=True
            )
        else:
            species[core_key] = get_core_interpolators(
                psi_norm, core_profiles, equilibrium, return3d=True
            )
    for edge_key, edge_profiles in edge_species.items():
        if edge_key not in core_species:
            species[edge_key] = get_edge_interpolators(grid, edge_profiles, b_field, return3d=True)

    for (species_type, charge), interp in species.items():
        if "velocity" not in interp or interp["velocity"] is None:
            interp["velocity"] = ConstantVector3D(Vector3D(0, 0, 0))

        if interp["density"] is None:
            print(f"Warning! Skipping {d['name']} species: density is not available.")
        if interp["temperature"] is None:
            print(f"Warning! Skipping {d['name']} species: temperature is not available.")

        distribution = Maxwellian(
            interp["density"],
            interp["temperature"],
            interp["velocity"],
            species_type.atomic_weight * atomic_mass,
        )

        plasma.composition.add(Species(species_type, charge, distribution))

    return plasma


def blend_core_edge_interpolators(
    core_interpolators: dict[str, Function3D | VectorFunction3D | None],
    edge_interpolators: dict[str, Function3D | VectorFunction3D | None],
    mask: Function2D | Function3D,
    return3d: bool = False,
) -> dict[str, Function3D | VectorFunction3D | None]:
    """Blend together interpolators for the core and edge using the modulating mask function.

    Parameters
    ----------
    core_interpolators : dict
        Dictionary with 2D or 3D core profiles interpolators.
    edge_interpolators : dict
        Dictionary with 2D or 3D edge profiles interpolators.
    mask : Function2D | Function3D
        Mask function used for blending: ``(1 - mask) * f_edge + mask * f_core``.
    return3d : bool, optional
        If True, return the 3D functions for 2D interpolators assuming
        rotational symmetry, by default False.

    Returns
    -------
    dict
        Dictionary with blended interpolators.
    """

    interpolators = {}

    for core_key, core_func in core_interpolators.items():
        edge_func = edge_interpolators[core_key] if core_key in edge_interpolators else None
        interpolators[core_key] = blend_core_edge_functions(core_func, edge_func, mask, return3d)

    for edge_key, edge_func in edge_interpolators.items():
        if edge_key not in core_interpolators:
            interpolators[edge_key] = blend_core_edge_functions(None, edge_func, mask, return3d)

    return interpolators


def blend_core_edge_functions(
    core_func: Function2D | Function3D | VectorFunction2D | VectorFunction3D | None,
    edge_func: Function2D | Function3D | VectorFunction2D | VectorFunction3D | None,
    mask: Function2D | Function3D,
    return3d: bool,
) -> Function3D | VectorFunction3D | None:
    """Blend together the core and edge interpolating functions using the modulating mask function.

    Parameters
    ----------
    core_func : Function2D | Function3D | VectorFunction2D | VectorFunction3D | None
        A 2D or 3D core interpolator.
    edge_func : Function2D | Function3D | VectorFunction2D | VectorFunction3D | None
        A 2D or 3D edge interpolator.
    mask : Function2D | Function3D
        The 2D or 3D mask function used for blending: (1 - mask) * f_edge + mask * f_core.
    return3d : bool, optional
        If True, return the 3D functions for 2D interpolators assuming
        rotational symmetry, by default False.

    Returns
    -------
    Function3D | VectorFunction3D | None
        Blended interpolator.
    """

    if core_func is None and edge_func is None:
        return None

    if core_func is not None and not isinstance(
        core_func, Function2D | Function3D | VectorFunction2D | VectorFunction3D
    ):
        raise ValueError("The core_func must be a 2D or 3D function.")

    if edge_func is not None and not isinstance(
        edge_func, Function2D | Function3D | VectorFunction2D | VectorFunction3D
    ):
        raise ValueError("The edge_func must be a 2D or 3D function.")

    if not isinstance(mask, Function2D | Function3D):
        raise ValueError("The mask must be a 2D or 3D function.")

    if core_func is None:
        if isinstance(edge_func, Function2D) and return3d:
            return AxisymmetricMapper(edge_func)

        if isinstance(edge_func, VectorFunction2D) and return3d:
            return VectorAxisymmetricMapper(edge_func)

        return edge_func

    if edge_func is None:
        if isinstance(core_func, Function2D) and return3d:
            return AxisymmetricMapper(core_func)

        if isinstance(core_func, VectorFunction2D) and return3d:
            return VectorAxisymmetricMapper(core_func)

        return core_func

    if (
        isinstance(core_func, Function2D)
        and isinstance(edge_func, Function2D)
        and isinstance(mask, Function2D)
    ):
        blended_func = Blend2D(edge_func, core_func, mask)
        return AxisymmetricMapper(blended_func) if return3d else blended_func

    if (
        isinstance(core_func, VectorFunction2D)
        and isinstance(edge_func, VectorFunction2D)
        and isinstance(mask, Function2D)
    ):
        blended_func = VectorBlend2D(edge_func, core_func, mask)
        return VectorAxisymmetricMapper(blended_func) if return3d else blended_func

    # unable to return 2D, convert to 3D

    if isinstance(core_func, Function2D):
        core_func = AxisymmetricMapper(core_func)

    if isinstance(core_func, VectorFunction2D):
        core_func = VectorAxisymmetricMapper(core_func)

    if isinstance(edge_func, Function2D):
        edge_func = AxisymmetricMapper(edge_func)

    if isinstance(edge_func, VectorFunction2D):
        edge_func = VectorAxisymmetricMapper(edge_func)

    if isinstance(mask, Function2D):
        mask = AxisymmetricMapper(mask)

    if isinstance(core_func, Function3D) and isinstance(edge_func, Function3D):
        return Blend3D(edge_func, core_func, mask)

    if isinstance(core_func, VectorFunction3D) and isinstance(edge_func, VectorFunction3D):
        return VectorBlend3D(edge_func, core_func, mask)

    raise ValueError("Cannot blend scalar and vector functions.")



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for common functions used to get IDS species information."""

from cherab.core.atomic.elements import Element, lookup_isotope
from imas.ids_defs import EMPTY_FLOAT, EMPTY_INT
from imas.ids_struct_array import IDSStructArray
from imas.ids_structure import IDSStructure

__all__ = [
    "get_ion_state",
    "get_neutral_state",
    "get_ion",
    "get_neutral",
    "get_element_list",
]


def get_ion_state(
    state: IDSStructure,
    state_index: int,
    elements: list[Element],
    grid_subset_index: int | None = None,
) -> tuple[str, frozenset]:
    """Get a unique identifier for an ion state.

    Parameters
    ----------
    state : IDSStructure
        The ion_state structure from IMAS.
    state_index : int
        Index of the state in the list of states.
    elements : list[`~cherab.core.atomic.elements.Element`]
        List of elements that make up the ion state.
    grid_subset_index : int, optional
        The grid subset index to use for 1D profiles, by default None.

    Returns
    -------
    species_type : str
        The type of species: 'ion', 'ion_bundle', 'molecule', or 'molecular_bundle'.
    species_id : frozenset
        A frozenset of key-value pairs that uniquely identify the species.
    """
    if state.z_min == EMPTY_FLOAT or state.z_max == EMPTY_FLOAT:
        if grid_subset_index is None:  # 1D profiles
            z_average = state.z_average_1d
        else:
            for s in state.z_average:
                if s.grid_subset_index == grid_subset_index:
                    z_average = s.values
        if len(z_average):  # probably, a bundle
            z_min = state.z_min.value if state.z_min != EMPTY_FLOAT else z_average.min()
            z_max = state.z_max.value if state.z_max != EMPTY_FLOAT else z_average.max()
        else:  # probably, a single ion
            z_min = state.z_min.value if state.z_min != EMPTY_FLOAT else state_index + 1
            z_max = state.z_max.value if state.z_max != EMPTY_FLOAT else z_min
    else:
        z_min = state.z_min.value
        z_max = state.z_max.value

    state_dict = {"name": state.name.strip()}

    if len(elements) > 1:  # molecular ions and bundles
        state_dict["elements"] = elements
        if z_min == z_max:
            species_type = "molecule"
            state_dict["z"] = z_min
            state_dict["electron_configuration"] = (
                str(state.electron_configuration) if len(state.electron_configuration) else None
            )
            state_dict["vibrational_mode"] = (
                str(state.vibrational_mode) if len(state.vibrational_mode) else None
            )
            state_dict["vibrational_level"] = (
                state.vibrational_level if state.vibrational_level != EMPTY_FLOAT else None
            )
        else:
            species_type = "molecular_bundle"
            state_dict["z_min"] = z_min
            state_dict["z_max"] = z_max
    else:  # ions and bundles
        state_dict["element"] = elements[0]
        if z_min == z_max:
            species_type = "ion"
            state_dict["z"] = z_min
            state_dict["electron_configuration"] = (
                str(state.electron_configuration) if len(state.electron_configuration) else None
            )
        else:
            species_type = "ion_bundle"
            state_dict["z_min"] = z_min
            state_dict["z_max"] = z_max
    species_id = frozenset(state_dict.items())

    return species_type, species_id


def get_neutral_state(state: IDSStructure, elements: list[Element]) -> tuple[str, frozenset]:
    """Get a unique identifier for a neutral state.

    Parameters
    ----------
    state : IDSStructure
        The neutral_state structure from IMAS.
    elements : list[`~cherab.core.atomic.elements.Element`]
        List of elements that make up the neutral state.

    Returns
    -------
    species_type : str
        The type of species: 'molecule' or 'ion'.
    species_id : frozenset
        A frozenset of key-value pairs that uniquely identify the species.
    """
    state_dict = {"name": state.name.strip()}

    if len(elements) > 1:  # molecules
        species_type = "molecule"
        state_dict["elements"] = elements
        state_dict["z"] = 0
        state_dict["electron_configuration"] = (
            str(state.electron_configuration) if len(state.electron_configuration) else None
        )
        state_dict["vibrational_mode"] = (
            str(state.vibrational_mode) if len(state.vibrational_mode) else None
        )
        state_dict["vibrational_level"] = (
            state.vibrational_level.value if state.vibrational_level != EMPTY_FLOAT else None
        )
    else:  # atoms
        species_type = "ion"
        state_dict["element"] = elements[0]
        state_dict["z"] = 0
        state_dict["electron_configuration"] = (
            str(state.electron_configuration) if len(state.electron_configuration) else None
        )

    species_id = frozenset(state_dict.items())

    return species_type, species_id


def get_ion(ion: IDSStructure, elements: list[Element]) -> tuple[str, frozenset]:
    """Get a unique identifier for an ion or molecule.

    Parameters
    ----------
    ion : IDSStructure
        The ion structure from IMAS.
    elements : list[`~cherab.core.atomic.elements.Element`]
        List of elements that make up the ion.

    Returns
    -------
    species_type : str
        The type of species: 'molecule' or 'ion'.
    species_id : frozenset
        A frozenset of key-value pairs that uniquely identify the species.
    """
    z_ion = int(ion.z_ion) if ion.z_ion != EMPTY_FLOAT else elements[0].atomic_number
    if len(elements) > 1:
        species_id = frozenset(
            {
                ("name", ion.name.strip()),
                ("elements", elements),
                ("z", z_ion),
                ("electron_configuration", None),
                ("vibrational_mode", None),
                ("vibrational_level", None),
            }
        )
        return "molecule", species_id

    species_id = frozenset(
        {
            ("name", ion.name.strip()),
            ("element", elements[0]),
            ("z", z_ion),
            ("electron_configuration", None),
        }
    )
    return "ion", species_id


def get_neutral(neutral, elements):
    """Get a unique identifier for a neutral or molecule.

    Parameters
    ----------
    neutral : IDSStructure
        The neutral structure from IMAS.
    elements : list[`~cherab.core.atomic.elements.Element`]
        List of elements that make up the neutral state.

    Returns
    -------
    species_type : str
        The type of species: 'molecule' or 'ion'.
    species_id : frozenset
        A frozenset of key-value pairs that uniquely identify the species.
    """
    if len(elements) > 1:
        species_id = frozenset(
            {
                ("name", neutral.name.strip()),
                ("elements", elements),
                ("z", 0),
                ("electron_configuration", None),
                ("vibrational_mode", None),
                ("vibrational_level", None),
            }
        )
        return "molecule", species_id

    species_id = frozenset(
        {
            ("name", neutral.name.strip()),
            ("element", elements[0]),
            ("z", 0),
            ("electron_configuration", None),
        }
    )
    return "ion", species_id


def get_element_list(element_aos: IDSStructArray) -> list[Element]:
    """Get a list of elements from an IDS element_aos structure.

    Parameters
    ----------
    element_aos : IDSStructArray
        The element_aos structure from IMAS.

    Returns
    -------
    list[`~cherab.core.atomic.elements.Element`]
        List of elements extracted from the element_aos structure.
    """
    elements = []
    for element in element_aos:
        mass_number = int(round(element.a))
        zn = int(round(element.z_n))
        isotope = lookup_isotope(zn, number=mass_number)
        if int(round(isotope.element.atomic_weight)) == mass_number:
            isotope = isotope.element  # prefer element over isotope
        atoms_n = 1 if element.atoms_n == EMPTY_INT else element.atoms_n.value
        for _ in range(atoms_n):
            elements.append(isotope)

    return elements



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for utility to load profiles."""

__all__ = ["warn_unsupported_species", "get_subset_name_index"]


def warn_unsupported_species(composition: dict[str, dict], species_type: str):
    """Warn if species of a given type are present in the composition dictionary.

    Parameters
    ----------
    composition : dict[str, dict]
        Dictionary with species composition.
    species_type : str
        Type of species to check for (e.g., 'ion_bundle', 'molecular_bundle').
    """
    if species_type in composition and len(composition[species_type]):
        print(
            f"Warning! Species of type '{species_type}' are currently not supported. "
            + "The following species will be skipped:"
        )
        names = []
        for species_id in composition[species_type]:
            d = {first: second for first, second in species_id}
            names.append(d["name"])
        print("; ".join(names))


def get_subset_name_index(subset_id_dict: dict, grid_subset_id: int | str) -> tuple[str, int]:
    """Get the name and index of a grid subset from its identifier.

    Parameters
    ----------
    subset_id_dict : dict
        Dictionary with grid subset indices.
    grid_subset_id : int | str
        Identifier of the grid subset. Either index or name.

    Returns
    -------
    grid_subset_name : str
        Name of the grid subset.
    grid_subset_index : int
        Index of the grid subset.
    """
    subset_id = subset_id_dict.copy()
    subset_id.update({value: key for key, value in subset_id.items()})

    try:
        grid_subset_index = int(grid_subset_id)
        grid_subset_name = subset_id[grid_subset_index]
    except ValueError as err1:
        try:
            grid_subset_name = str(grid_subset_id)
            grid_subset_index = subset_id[grid_subset_name]
        except KeyError:
            raise ValueError(f"Unable to find a grid subset with ID {grid_subset_id}.") from err1
    except KeyError as err2:
        raise ValueError(f"Unable to find a grid subset with ID {grid_subset_id}.") from err2

    return grid_subset_name, grid_subset_index



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading unstructured 2D grids from IMAS grid_ggd IDS structure."""

import numpy as np

from cherab.imas.ggd import UnstructGrid2D
from imas.ids_defs import EMPTY_INT
from imas.ids_structure import IDSStructure

__all__ = ["load_unstruct_grid_2d"]

VERTEX_DIMENSION = 0
EDGE_DIMENSION = 1
FACE_DIMENSION = 2


def load_unstruct_grid_2d(grid_ggd: IDSStructure, space_index=0, with_subsets=False):
    """Load unstructured 2D grid from the grid_ggd structure.

    Parameters
    ----------
    grid_ggd : IDSStructure
        The grid_ggd structure.
    space_index : int, optional
        The index of the grid space, by default 0.
    with_subsets : bool, optional
        Read grid subset data, by default False.

    Returns
    -------
    grid : UnstructGrid2D
        Unstructured 2D grid object.
    subsets : dict, optional
        Dictionary with grid subsets for each subset name containing the indices of the cells from
        that subset. Note that 'Cells' subset is included only if cell indices are specified.
    subset_id : dict, optional
        Dictionary with grid subset indices.
    """

    space = grid_ggd.space[space_index]

    # Check if the grid is 2D
    if len(space.objects_per_dimension) != 3:
        raise ValueError("The load_unstruct_grid_2d() supports only unstructured 2D grids.")

    grid_name = str(grid_ggd.identifier.name)

    # Reading vertices
    num_vert = len(space.objects_per_dimension[VERTEX_DIMENSION].object)
    vertices = np.empty((num_vert, 2), dtype=np.float64)
    for i in range(num_vert):
        vertices[i] = space.objects_per_dimension[VERTEX_DIMENSION].object[i].geometry[:2]

    # Reading polygonal cells
    cells = []
    winding_ok = True
    for object in space.objects_per_dimension[FACE_DIMENSION].object:
        # getting cell from nodes
        cell = np.asarray_chkfinite(object.nodes, dtype=np.int32) - 1  # Fortran to C indexing
        if cell.size > 3:
            # trying to get the nodes in winding order by parsing the edges
            edge_dict = {}
            for boundary in object.boundary:
                n1, n2 = (
                    space.objects_per_dimension[EDGE_DIMENSION].object[boundary.index - 1].nodes - 1
                )  # Fortran to C indexing
                if n1 not in cell or n2 not in cell:  # fail, error in the data
                    edge_dict = {}
                    break
                if n1 in edge_dict:
                    if n2 == edge_dict[n1][0]:  # fail, error in the data
                        edge_dict = {}
                        break
                    edge_dict[n1][1] = n2
                else:
                    edge_dict[n1] = [n2, -1]
                if n2 in edge_dict:
                    if n1 == edge_dict[n2][0]:  # fail, error in the data
                        edge_dict = {}
                        break
                    edge_dict[n2][1] = n1
                else:
                    edge_dict[n2] = [n1, -1]
            if len(edge_dict) == cell.size:  # success, getting the cell nodes in winding order
                cell1 = np.empty(len(edge_dict), dtype=np.int32)
                cell1[0] = cell[0]
                pair = edge_dict[cell1[0]]
                cell1[1] = cell[1] if cell[1] in pair else pair[0]
                for i in range(2, cell1.size):
                    pair = edge_dict[cell1[i - 1]]
                    cell1[i] = pair[1] if cell1[i - 2] == pair[0] else pair[0]
                cell = cell1
            else:
                winding_ok = False

        cells.append(cell)

    if not winding_ok:
        print("Warning! Unable to verify that the cell nodes are in the winging order.")

    grid = UnstructGrid2D(vertices, cells, name=grid_name)

    if not with_subsets:
        return grid

    # Reading grid subsets (2D only)
    cell_subset_ids = (5, 22, 23, 24, 25, 38, 39, 40)
    subsets = {}
    subset_id = {}
    for subset in grid_ggd.grid_subset:
        dimension_is_2d = subset.dimension == FACE_DIMENSION + 1  # C to Fortran indexing
        known_subset_id = (
            subset.dimension != EMPTY_INT and subset.identifier.index in cell_subset_ids
        )
        if (dimension_is_2d or known_subset_id) and len(subset.element):
            name = str(subset.identifier.name)
            indices = np.empty(len(subset.element), dtype=np.int32)
            for i, element in enumerate(subset.element):
                if len(element.object) > 1:
                    print(
                        f"Warning! Skipping grid subset {name}, "
                        + "because it includes cells not present in the original grid."
                    )
                    break
                indices[i] = element.object[0].index.value
            subsets[name] = indices - 1  # Fortran to C indexing
            subset_id[name] = subset.identifier.index.value

    return grid, subsets, subset_id



import pytest
from raysect.core.scenegraph import World
from raysect.primitive import Subtract

from cherab.core import Maxwellian, Plasma
from cherab.imas.plasma import load_plasma
from cherab.imas.plasma.equilibrium import load_magnetic_field


def test_load_plasma_basic(path_iter_jintrac: str):
    """Test basic loading of plasma data from an IMAS file."""
    plasma = load_plasma(path_iter_jintrac, "r")

    # Test that a Plasma object is returned
    assert isinstance(plasma, Plasma)


def test_load_plasma_geometry(path_iter_jintrac: str):
    """Test that the plasma geometry is properly set."""
    plasma = load_plasma(path_iter_jintrac, "r")

    # Check that geometry is set and has the expected type
    assert plasma.geometry is not None
    assert isinstance(plasma.geometry, Subtract)

    # Check that geometry transform is set
    assert plasma.geometry_transform is not None


def test_load_plasma_electron_distribution(path_iter_jintrac: str):
    """Test that electron distribution is properly loaded."""
    plasma = load_plasma(path_iter_jintrac, "r")

    # Check that electron distribution exists
    assert plasma.electron_distribution is not None
    assert isinstance(plasma.electron_distribution, Maxwellian)


def test_load_plasma_with_time(path_iter_jintrac: str):
    """Test loading plasma with specific time parameter."""

    # Test with default time (should not raise error)
    plasma1 = load_plasma(path_iter_jintrac, "r", time=0.0)
    assert isinstance(plasma1, Plasma)

    # we typically have time=450 in the test dataset
    plasma2 = load_plasma(path_iter_jintrac, "r", time=450.0)
    assert isinstance(plasma2, Plasma)


def test_load_plasma_with_occurrence(path_iter_jintrac: str):
    """Test loading plasma with different occurrence values."""

    # Test with default occurrence core
    plasma1 = load_plasma(path_iter_jintrac, "r", occurrence_core=0)
    assert isinstance(plasma1, Plasma)

    # Test with default occurrence edge
    plasma2 = load_plasma(path_iter_jintrac, "r", occurrence_edge=0)
    assert isinstance(plasma2, Plasma)

    # Test with default occurrence core and edge
    plasma3 = load_plasma(path_iter_jintrac, "r", occurrence_core=0, occurrence_edge=0)
    assert isinstance(plasma3, Plasma)


def test_load_plasma_with_parent(path_iter_jintrac: str):
    """Test loading plasma with a parent scene graph node."""
    world = World()

    plasma = load_plasma(path_iter_jintrac, "r", parent=world)
    assert isinstance(plasma, Plasma)
    assert plasma.parent is world


def test_load_plasma_with_magnetic_field(path_iter_jintrac: str):
    """Test loading plasma with external magnetic field."""

    # Load magnetic field separately
    try:
        b_field = load_magnetic_field(path_iter_jintrac, "r")
        plasma = load_plasma(path_iter_jintrac, "r", b_field=b_field)
        assert isinstance(plasma, Plasma)
        assert plasma.b_field is not None
    except RuntimeError:
        # If magnetic field loading fails, test should still pass
        # as this is handled gracefully in the function
        pass


def test_load_plasma_time_threshold(path_iter_jintrac: str):
    """Test loading plasma with time threshold parameter."""

    # Test with large time threshold (should work)
    plasma = load_plasma(path_iter_jintrac, "r", time_threshold=500)
    assert isinstance(plasma, Plasma)

    # Test with very small time threshold (should raise error)
    with pytest.raises(RuntimeError):
        load_plasma(path_iter_jintrac, "r", time_threshold=0.0)


def test_load_plasma_composition(path_iter_jintrac: str):
    """Test that plasma composition contains expected species."""
    plasma = load_plasma(path_iter_jintrac, "r")

    # Check that we have some ion species (exact species depend on data)
    assert len(plasma.composition) > 0


def test_load_plasma_error_handling(path_iter_jintrac: str):
    """Test error handling for invalid inputs."""
    # Test with non-existent file path
    with pytest.raises((FileNotFoundError, RuntimeError, OSError)):
        load_plasma("non_existent_path", "r")

    # Test with invalid mode
    with pytest.raises((ValueError, RuntimeError, OSError)):
        load_plasma(path_iter_jintrac, "invalid_mode")



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading plasma equilibrium and magnetic field from the equilibrium IDS."""

import numpy as np
from raysect.core.math.function.float import Interpolator1DArray, Interpolator2DArray
from raysect.core.math.function.vector3d import FloatToVector3DFunction2D

from cherab.tools.equilibrium import EFITEquilibrium
from imas import DBEntry

from ..ids.common import get_ids_time_slice
from ..ids.equilibrium import load_equilibrium_data, load_magnetic_field_data

__all__ = ["load_equilibrium", "load_magnetic_field"]


def load_equilibrium(
    *args,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = np.inf,
    with_psi_interpolator: bool = False,
    **kwargs,
) -> tuple[EFITEquilibrium, Interpolator1DArray | None] | EFITEquilibrium:
    """Load plasma equilibrium from the equilibrium IDS and create an `EFITEquilibrium` object.

    Parameters
    ----------
    *args
        Arguments passed to `~imas.db_entry.DBEntry`.
    time : float, optional
        Time moment, by default 0.
    occurrence : int, optional
        Instance index of the 'equilibrium' IDS, by default 0.
    time_threshold : float, optional
        Maximum allowable difference between the specified time and the nearest available
        time, by default `numpy.inf`.
    with_psi_interpolator : bool, optional
        If True, returns the ``psi_norm(rho_tor_norm)`` interpolator; otherwise, returns only the
        equilibrium object.
    **kwargs
        Keyword arguments passed to `~imas.db_entry.DBEntry`.

    Returns
    -------
    equilibrium : EFITEquilibrium
        The `EFITEquilibrium` object.
    psi_interpolator : Function1D, optional
        If ``with_psi_interpolator`` is True and ``rho_tor_norm`` is available, returns the
        ``psi_norm(rho_tor_norm)`` interpolator.
        If rho_tor_norm is not available, returns None.
        Otherwise, returns only the equilibrium object.
    """
    with DBEntry(*args, **kwargs) as entry:
        equilibrium_ids = get_ids_time_slice(
            entry, "equilibrium", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    equilibrium_dict = load_equilibrium_data(equilibrium_ids)

    cocos_11to3(equilibrium_dict)

    equilibrium_dict["psi_norm"][0] = min(0, equilibrium_dict["psi_norm"][0])
    equilibrium_dict["psi_norm"][-1] = max(1.0, equilibrium_dict["psi_norm"][-1])

    f_profile = np.array([equilibrium_dict["psi_norm"], equilibrium_dict["f"]])
    q_profile = np.array([equilibrium_dict["psi_norm"], equilibrium_dict["q"]])

    equilibrium = EFITEquilibrium(
        equilibrium_dict["r"],
        equilibrium_dict["z"],
        equilibrium_dict["psi_grid"],
        equilibrium_dict["psi_axis"],
        equilibrium_dict["psi_lcfs"],
        equilibrium_dict["magnetic_axis"],
        equilibrium_dict["x_points"],
        equilibrium_dict["strike_points"],
        f_profile,
        q_profile,
        equilibrium_dict["b_vacuum_radius"],
        equilibrium_dict["b_vacuum_magnitude"],
        equilibrium_dict["lcfs_polygon"],
        None,
        equilibrium_dict["time"],
    )

    if not with_psi_interpolator:
        return equilibrium

    if equilibrium_dict["rho_tor_norm"] is None:
        return equilibrium, None

    psi_interpolator = Interpolator1DArray(
        equilibrium_dict["rho_tor_norm"], equilibrium_dict["psi_norm"], "cubic", "none", 0
    )

    return equilibrium, psi_interpolator


def load_magnetic_field(
    *args,
    time: float = 0,
    occurrence: int = 0,
    time_threshold: float = np.inf,
    **kwargs,
) -> FloatToVector3DFunction2D:
    """Load the magnetic field from the equilibrium IDS and returns a VectorFunction2D interpolator.

    Parameters
    ----------
    *args
        Arguments passed to `~imas.db_entry.DBEntry`.
    time : float, optional
        Time moment, by default 0.
    occurrence : int, optional
        Instance index of the 'equilibrium' IDS, by default 0.
    time_threshold : float, optional
        Sets the maximum allowable difference between the specified time and the nearest available
        time, by default `numpy.inf`.
    **kwargs
        Keyword arguments passed to `~imas.db_entry.DBEntry`.

    Returns
    -------
    VectorFunction2D
        The magnetic field interpolator.
    """

    with DBEntry(*args, **kwargs) as entry:
        equilibrium_ids = get_ids_time_slice(
            entry, "equilibrium", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    if not len(equilibrium_ids.time_slice):
        raise RuntimeError("Equilibrium IDS does not have a time slice.")

    b_dict = load_magnetic_field_data(equilibrium_ids.time_slice[0].profiles_2d)

    br = Interpolator2DArray(b_dict["r"], b_dict["z"], b_dict["b_field_r"], "cubic", "none", 0, 0)
    btor = Interpolator2DArray(
        b_dict["r"], b_dict["z"], b_dict["b_field_phi"], "cubic", "none", 0, 0
    )
    bz = Interpolator2DArray(b_dict["r"], b_dict["z"], b_dict["b_field_z"], "cubic", "none", 0, 0)

    return FloatToVector3DFunction2D(br, btor, bz)


def cocos_11to3(equilibrium_dict: dict) -> None:
    """Convert from COCOS 11 convention used in IMAS to COCOS 3 convention used in EFIT.

    This function modifies the equilibrium_dict in place to convert the coordinates
    and other relevant data from the COCOS 11 convention to the COCOS 3 convention.

    Parameters
    ----------
    equilibrium_dict : dict
        The equilibrium data dictionary to modify in place.
    """

    equilibrium_dict["psi_grid"] = -equilibrium_dict["psi_grid"] / (2.0 * np.pi)
    equilibrium_dict["psi_axis"] = -equilibrium_dict["psi_axis"] / (2.0 * np.pi)
    equilibrium_dict["psi_lcfs"] = -equilibrium_dict["psi_lcfs"] / (2.0 * np.pi)
    equilibrium_dict["q"] = -equilibrium_dict["q"]
    if equilibrium_dict["phi"] is not None:
        equilibrium_dict["phi"] = -equilibrium_dict["phi"]



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading unstructured 3D grids from IMAS grid_ggd IDS structure."""

import numpy as np

from imas.ids_structure import IDSStructure

from ....ggd import UnstructGrid2DExtended

__all__ = ["load_unstruct_grid_2d_extended", "load_unstruct_grid_3d"]

# Constants for space indices and dimensions
SPACE_RZ = 0
SPACE_FOURIER = 1
VERTEX_DIMENSION = 0
EDGE_DIMENSION = 1
FACE_DIMENSION = 2
TETRA_IN_CELL = 5
NUM_TOROIDAL = 64


def load_unstruct_grid_2d_extended(
    grid_ggd: IDSStructure, with_subsets: bool = False, num_toroidal: int | None = None
):
    """Load unstructured 2D grid extended in 3D from the grid_ggd structure.

    Parameters
    ----------
    grid_ggd : IDSStructure
        The grid_ggd structure.
    with_subsets : bool, optional
        Read grid subset data if True, by default False.
    num_toroidal : int, optional
        Number of toroidal points, by default None.

    Returns
    -------
    UnstructGrid2DExtended
        The unstructured 2D grid extended in 3D.
    """
    # Check if the number of toroidal points is specified
    if isinstance(num_toroidal, int):
        if num_toroidal < 1:
            raise ValueError("The number of toroidal points must be greater than 0.")
        num_toroidal = num_toroidal
    else:
        num_toroidal = NUM_TOROIDAL

    # Get the R-Z space
    space = grid_ggd.space[SPACE_RZ]

    # Check if the grid is 2D
    if len(space.objects_per_dimension) != 3:
        raise ValueError(
            "The load_unstruct_grid_2d_extended() supports only unstructured extended 2D grids."
        )

    grid_name = grid_ggd.identifier.name

    # =========================================
    # Reading vertices (poloidal and toroidal)
    # =========================================
    num_poloidal = len(space.objects_per_dimension[VERTEX_DIMENSION].object)
    num_vert = num_poloidal * num_toroidal
    vertices_rpz = np.empty((num_vert, 3), dtype=float)

    for i_phi, i_pol in np.ndindex(num_toroidal, num_poloidal):
        phi = 2.0 * np.pi * i_phi / num_toroidal
        vertices_rpz[i_pol + i_phi * num_poloidal] = (
            space.objects_per_dimension[VERTEX_DIMENSION].object[i_pol].geometry[0],
            phi,
            space.objects_per_dimension[VERTEX_DIMENSION].object[i_pol].geometry[1],
        )
    # Convert to cartesian coordinates
    vertices = np.empty((num_vert, 3), dtype=float)
    p = vertices_rpz[:, 1]
    vertices[:, 0] = vertices_rpz[:, 0] * np.cos(p)
    vertices[:, 1] = vertices_rpz[:, 0] * np.sin(p)
    vertices[:, 2] = vertices_rpz[:, 2]

    # =========================================
    # Reading cells indices
    # =========================================
    faces = space.objects_per_dimension[FACE_DIMENSION].object
    num_faces = len(faces)
    cells = np.zeros((num_faces * num_toroidal, 8), dtype=np.int32)
    i_cell = 0
    for i_phi, i_face in np.ndindex(num_toroidal, num_faces):
        i_phi_next = i_phi + 1 if i_phi + 1 < num_toroidal else 0
        cells[i_cell] = np.hstack(
            [
                (faces[i_face].nodes - 1) + i_phi * num_poloidal,
                (faces[i_face].nodes - 1) + i_phi_next * num_poloidal,
            ]
        )
        i_cell += 1

    grid = UnstructGrid2DExtended(
        vertices, cells, num_faces, num_poloidal, num_toroidal, name=grid_name
    )

    if not with_subsets:
        return grid
    else:
        raise NotImplementedError("Reading grid subsets is not implemented yet.")


def load_unstruct_grid_3d(grid_ggd: IDSStructure, space_index: int = 0, with_subsets: bool = False):
    """Load unstructured 3D grid from the grid_ggd structure.

    .. warning::
        This function is a placeholder for future implementation.

    Parameters
    ----------
    grid_ggd : IDSStructure
        The grid_ggd structure.
    space_index : int, optional
        The index of the space to read, by default 0.
    with_subsets : bool, optional
        Read grid subset data if True, by default False.

    Returns
    -------
    UnstructGrid3D
        The unstructured 3D grid.
    """
    raise NotImplementedError("Loading unstructured 3D grids will be implemented in the future.")



from matplotlib import pyplot as plt
from raysect.core.math.function.vector3d import FloatToVector3DFunction2D

from cherab.imas.plasma.equilibrium import load_equilibrium, load_magnetic_field
from cherab.tools.equilibrium import EFITEquilibrium, plot_equilibrium

plt.rcParams["backend"] = "Agg"  # Use non-interactive backend for testing


def test_load_equilibrium(path_iter_jintrac: str):
    """Test loading of equilibrium data from an IMAS file."""
    equilibrium = load_equilibrium(path_iter_jintrac, "r")

    # Test that equilibrium object is returned
    assert isinstance(equilibrium, EFITEquilibrium)

    # Test plotting function
    plot_equilibrium(equilibrium)


def test_load_magnetic_field(path_iter_jintrac: str):
    """Test loading of magnetic field data from an IMAS file."""
    magnetic_field = load_magnetic_field(path_iter_jintrac, "r")

    # Test that Vector3DFunction2D object is returned
    assert isinstance(magnetic_field, FloatToVector3DFunction2D)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading edge plasma profiles from the edge_profiles IDS."""

import numpy as np
from raysect.core.math import Vector3D, translate
from raysect.core.math.function.float import Constant2D, Constant3D, Function2D, Function3D
from raysect.core.math.function.vector3d import Constant2D as ConstantVector2D
from raysect.core.math.function.vector3d import Constant3D as ConstantVector3D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.primitive import Cylinder, Subtract
from scipy.constants import atomic_mass, electron_mass

from cherab.core import Maxwellian, Plasma, Species
from cherab.core.math import AxisymmetricMapper, VectorAxisymmetricMapper
from cherab.core.utility import RecursiveDict
from cherab.tools.equilibrium.efit import FluxSurfaceNormal, PoloidalFieldVector
from imas import DBEntry
from imas.ids_structure import IDSStructure

from ..ggd.base_mesh import GGDGrid
from ..ids.common import get_ids_time_slice
from ..ids.common.ggd import load_grid
from ..ids.edge_profiles import load_edge_species
from ..math import UnitVector2D
from .equilibrium import load_equilibrium, load_magnetic_field
from .utility import get_subset_name_index, warn_unsupported_species

__all__ = ["load_edge_plasma"]


def load_edge_plasma(
    *args,
    time: float = 0,
    occurrence: int = 0,
    grid_ggd: IDSStructure | None = None,
    grid_subset_id: int | str = 5,
    b_field: VectorFunction2D | None = None,
    time_threshold: float = np.inf,
    parent: _NodeBase | None = None,
    **kwargs,
) -> Plasma:
    """Load edge profiles and Create a `~cherab.core.plasma.node.Plasma` object.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time : float, optional
        Time moment for the edge plasma, by default 0.
    occurrence : int, optional
        Instance index of the 'edge_profiles' IDS, by default 0.
    grid_ggd : IDSStructure, optional
        Alternative grid_ggd structure with the grid description, by default None.
    grid_subset_id : int | str, optional
        Identifier of the grid subset. Either index or name, by default 5 ("Cells").
    b_field : VectorFunction2D, optional
        Alternative 2D interpolator of the magnetic field vector (Br, Btor, Bz).
        Default is None. The magnetic field will be loaded from the 'equilibrium' IDS.
    time_threshold : float, optional
        Sets the maximum allowable difference between the specified time and the nearest
        available time, by default `numpy.inf`.
    parent : _NodeBase, optional
        Parent node in the Raysect scene-graph, by default None.
        Normally, `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `~cherab.core.plasma.node.Plasma`
        Plasma object with edge profiles.
    """

    with DBEntry(*args, **kwargs) as entry:
        edge_profiles_ids = get_ids_time_slice(
            entry, "edge_profiles", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    if not len(edge_profiles_ids.grid_ggd) and grid_ggd is None:
        raise RuntimeError(
            "The 'grid_ggd' AOS of the edge_profiles IDS is empty "
            + "and an alternative grid_ggd structure is not provided."
        )

    if not len(edge_profiles_ids.ggd):
        raise RuntimeError("The 'ggd' AOS of the edge_profiles IDS is empty.")

    grid_ggd = grid_ggd or edge_profiles_ids.grid_ggd[0]
    grid, subsets, subset_id = load_grid(grid_ggd, with_subsets=True)

    grid_subset_name, grid_subset_index = get_subset_name_index(subset_id, grid_subset_id)

    if np.all(subsets[grid_subset_name] != np.arange(grid.num_cell, dtype=int)):
        # To reduce memory usage, create the sub-grid only if needed.
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)

    composition = load_edge_species(edge_profiles_ids.ggd[0], grid_subset_index=grid_subset_index)

    name = f"IMAS edge plasma: time {edge_profiles_ids.time[0]}, uri {entry.uri}."
    plasma = Plasma(parent=parent, name=name)

    # Create plasma geometry
    radius_outer = grid.mesh_extent["rmax"]
    radius_inner = grid.mesh_extent["rmin"]
    height = grid.mesh_extent["zmax"] - grid.mesh_extent["zmin"]
    zmin = grid.mesh_extent["zmin"]
    plasma.geometry = Subtract(Cylinder(radius_outer, height), Cylinder(radius_inner, height))
    plasma.geometry_transform = translate(0, 0, zmin)

    if b_field is None:
        try:
            b_field = load_magnetic_field(*args, time=time, occurrence=occurrence, **kwargs)
        except RuntimeError:
            try:
                b_field = load_equilibrium(
                    *args, time=time, occurrence=occurrence, **kwargs
                ).b_field
            except RuntimeError:
                print("Warning! No magnetic field data available in the equilibrium IDS.")

    if b_field is not None:
        plasma.b_field = VectorAxisymmetricMapper(b_field)

    # Add electron species
    electrons = get_edge_interpolators(grid, composition["electron"], b_field, return3d=True)

    if electrons["density"] is None:
        print("Unable to create Edge Plasma: electron density is not available.")
    if electrons["temperature"] is None:
        print("Unable to create Edge Plasma: electron temperature is not available.")

    plasma.electron_distribution = Maxwellian(
        electrons["density"], electrons["temperature"], electrons["velocity"], electron_mass
    )

    warn_unsupported_species(composition, "molecule")
    warn_unsupported_species(composition, "molecular_bundle")
    warn_unsupported_species(composition, "ion_bundle")

    # Add ion and neutral species
    for species_id, profiles in composition["ion"].items():
        d = {first: second for first, second in species_id}
        species_type = d["element"]
        charge = int(round(d["z"]))

        sp_key = (species_type, charge)
        if sp_key in plasma.composition:
            print(
                f"Warning! Skipping {d['name']} species. "
                + f"Species with the same (element, charge): {sp_key} is already added."
            )
            continue

        interp = get_edge_interpolators(grid, profiles, b_field, return3d=True)

        if interp["density"] is None:
            print(f"Warning! Skipping {d['name']} species: density is not available.")
        if interp["temperature"] is None:
            print(f"Warning! Skipping {d['name']} species: temperature is not available.")

        distribution = Maxwellian(
            interp["density"],
            interp["temperature"],
            interp["velocity"],
            species_type.atomic_weight * atomic_mass,
        )

        plasma.composition.add(Species(species_type, charge, distribution))

    return plasma


def get_edge_interpolators(
    grid: GGDGrid,
    profiles: dict[str, np.ndarray | None],
    b_field: VectorFunction2D | None = None,
    return3d: bool = False,
) -> dict[str, Function3D | Function2D | VectorFunction2D | VectorFunction2D | None]:
    """Create interpolators for the profiles defined on a grid.

    Parameters
    ----------
    grid : GGDGrid
        GGD-compatible grid object.
    profiles : dict[str, np.ndarray | None]
        Dictionary with edge plasma profiles.
    b_field : VectorFunction2D, optional
        2D interpolator of the magnetic field vector (Br, Btor, Bz), by default None.
    return3d : bool, optional
        If True, return the 3D functions for 2D interpolators assuming
        rotational symmetry, by default False.

    Returns
    -------
    dict[str, Function3D | Function2D | None]
        Dictionary with edge interpolators.
    """

    interpolators = RecursiveDict()

    for prof_key, profile in profiles.items():
        if "velocity" in prof_key:
            continue
        if profile is not None:
            func = grid.interpolator(profile)
            if isinstance(func, Function2D) and return3d:
                func = AxisymmetricMapper(func)
            interpolators[prof_key] = func
        else:
            interpolators[prof_key] = None

    vector_func = _get_velocity_interpolators(grid, profiles, b_field)
    if isinstance(vector_func, VectorFunction2D) and return3d:
        vector_func = VectorAxisymmetricMapper(vector_func)
    interpolators["velocity"] = vector_func

    return interpolators.freeze()


def _get_velocity_interpolators(grid, profiles, b_field=None):
    # Note: np.all(None == 0) returns False
    vrad = None if np.all(profiles["velocity_radial"] == 0) else profiles["velocity_radial"]
    vpol = None if np.all(profiles["velocity_poloidal"] == 0) else profiles["velocity_poloidal"]
    vpar = None if np.all(profiles["velocity_parallel"] == 0) else profiles["velocity_parallel"]
    vtor = None if np.all(profiles["velocity_phi"] == 0) else profiles["velocity_phi"]
    vr = None if np.all(profiles["velocity_r"] == 0) else profiles["velocity_r"]
    vz = None if np.all(profiles["velocity_z"] == 0) else profiles["velocity_z"]

    if not b_field:
        return _get_cylindrical_velocity_interpolators(grid, vr, vz, vtor)

    if vrad is None and vr is not None and vz is not None:
        if vtor is None and vpar is not None:
            _, vtor = _get_components_from_vpar(grid, vpar, b_field)
        return _get_cylindrical_velocity_interpolators(grid, vr, vz, vtor)

    if vpar is None:
        return _get_poloidal_velocity_interpolators(grid, vpol, vrad, vtor, b_field)

    return _get_parallel_velocity_interpolators(grid, vpar, vrad, b_field)


def _get_cylindrical_velocity_interpolators(grid, vr, vz, vtor):
    if vr is None and vz is None and vtor is None:
        if grid.dimension == 2:
            return ConstantVector2D(
                Vector3D(0, 1.0e-16, 0)
            )  # avoid zero-length vectors for blending

        return ConstantVector3D(Vector3D(0, 1.0e-16, 0))  # avoid zero-length vectors for blending

    if vr is None:
        vr = np.zeros(grid.num_cell, dtype=np.float64)
    if vz is None:
        vz = np.zeros(grid.num_cell, dtype=np.float64)
    if vtor is None:
        vtor = np.zeros(grid.num_cell, dtype=np.float64)

    return grid.vector_interpolator(np.array([vr, vtor, vz]))


def _get_parallel_velocity_interpolators(grid, vpar, vrad, b_field):
    if vpar is None and vrad is None:
        if grid.dimension == 2:  # 2D case
            return ConstantVector2D(
                Vector3D(0, 1.0e-16, 0)
            )  # avoid zero-length vectors for blending

        return ConstantVector3D(Vector3D(0, 1.0e-16, 0))  # avoid zero-length vectors for blending

    const_func = Constant2D if grid.dimension == 2 else Constant3D

    vpar_i = const_func(0) if vpar is None else grid.interpolator(vpar)
    vrad_i = const_func(0) if vrad is None else grid.interpolator(vrad)

    parallel_vector = UnitVector2D(b_field)
    surface_normal = FluxSurfaceNormal(b_field)

    if grid.dimension == 3:  # 3D case
        parallel_vector = VectorAxisymmetricMapper(parallel_vector)
        surface_normal = VectorAxisymmetricMapper(surface_normal)

    return vpar_i * parallel_vector + vrad_i * surface_normal


def _get_poloidal_velocity_interpolators(grid, vpol, vrad, vtor, b_field):
    if vpol is None and vrad is None and vtor is None:
        if grid.dimension == 2:  # 2D case
            return ConstantVector2D(
                Vector3D(0, 1.0e-16, 0)
            )  # avoid zero-length vectors for blending

        return ConstantVector3D(Vector3D(0, 1.0e-16, 0))  # avoid zero-length vectors for blending

    const_func = Constant2D if grid.dimension == 2 else Constant3D

    vpol_i = const_func(0) if vpol is None else grid.interpolator(vpol)
    vrad_i = const_func(0) if vrad is None else grid.interpolator(vrad)
    vtor_i = const_func(0) if vtor is None else grid.interpolator(vtor)

    poloidal_vector = PoloidalFieldVector(b_field)
    surface_normal = FluxSurfaceNormal(b_field)
    toroidal_vector = ConstantVector2D(Vector3D(0, 1, 0))

    if grid.dimension == 3:  # 3D case
        poloidal_vector = VectorAxisymmetricMapper(poloidal_vector)
        surface_normal = VectorAxisymmetricMapper(surface_normal)
        toroidal_vector = VectorAxisymmetricMapper(toroidal_vector)

    return vpol_i * poloidal_vector + vrad_i * surface_normal + vtor_i * toroidal_vector


def _get_components_from_vpar(grid, vpar, b_field):
    vpol = np.zeros(grid.num_cell, dtype=np.float64)
    vtor = np.zeros(grid.num_cell, dtype=np.float64)

    for i, cell_centre in enumerate(grid.cell_centre):
        if grid.dimension == 2:  # 2D case
            r, z = cell_centre
        else:  # 3D case
            if grid.coordinate_system == "cartesian":
                r = np.sqrt(cell_centre[0] ** 2 + cell_centre[1] ** 2)
                z = cell_centre[2]
            else:
                r, _, z = cell_centre
        try:
            b_field = b_field(r, z)
            vpol[i] = np.sqrt(b_field.x**2 + b_field.z**2) * (vpar[i] / b_field.length)
            vtor[i] = vpar[i] * b_field.y / b_field.length
        except ValueError:  # Outside equilibrium grid
            continue

    return vpol, vtor



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading core-profile-related data from IMAS IDS structures."""

import numpy as np

from imas.ids_structure import IDSStructure

from ..common.species import (
    get_element_list,
    get_ion,
    get_ion_state,
    get_neutral,
    get_neutral_state,
)

__all__ = ["load_core_profiles", "load_core_grid", "load_core_species"]


def load_core_profiles(
    species_struct: IDSStructure, backup_species_struct: IDSStructure | None = None
) -> dict[str, np.ndarray | None]:
    """Load core profiles from a given species structure.

    The returned dictionary values for missing profiles are None.

    Parameters
    ----------
    species_struct : IDSStructure
        IDS structure containing the profiles for a single species.
    backup_species_struct : IDSStructure | None, optional
        The backup ids structure that is used if the profile is missing in species_struct,
        by default None.

    Returns
    -------
    dict[str, ndarray | None]
        Dictionary with the profiles: ``density``, ``density_thermal``, ``density_fast``,
        ``temperature`` and ``z_average_1d``.
    """

    profiles = {
        "density": None,
        "density_thermal": None,
        "density_fast": None,
        "temperature": None,
        "z_average_1d": None,
    }

    for name in profiles:
        profiles[name] = _get_profile(species_struct, name)
        if profiles[name] is None and backup_species_struct is not None:
            profiles[name] = _get_profile(backup_species_struct, name)

    return profiles


def load_core_grid(grid_struct: IDSStructure) -> dict[str, np.ndarray | None]:
    """Load grid properties of the core profiles.

    The returned dictionary values for missing data are None.

    Parameters
    ----------
    grid_struct : IDSStructure
        The IDS structure containing the grid data for 1D profiles.

    Returns
    -------
    dict[str, np.ndarray | None]
        A dictionary with the following keys: ``rho_tor_norm``, ``psi``, ``volume``,
        ``area``, ``surface``.
    """

    grid = {
        "rho_tor_norm": None,
        "volume": None,
        "area": None,
        "surface": None,
        "psi": None,
    }

    for name in grid:
        grid[name] = _get_profile(grid_struct, name)

    return grid


def load_core_species(profiles_struct: IDSStructure) -> dict[str, dict[str, np.ndarray | None]]:
    """Load core plasma species and their profiles from a given profiles IDS structure.

    The returned dictionary has the following structure.

    .. autolink-skip::
    .. code-block:: python

        {
            'electron': {
                'density': array,
                'temperature': array,
                ...
            },
            'molecule': {
                molecule_id: {  # frozenset identifier
                    'density': array,
                    'temperature': array,
                    ...
                },
                ...
            'molecular_bundle': {
                molecular_bundle_id: {  # frozenset identifier
                    'density': array,
                    'temperature': array,
                    ...
                },
                ...
            'ion': {
                ion_id: {  # frozenset identifier
                    'density': array,
                    'temperature': array,
                    ...
                },
                ...
            'ion_bundle': {
                ion_bundle_id: {  # frozenset identifier
                    'density': array,
                    'temperature': array,
                    ...
                },
            },
        }

    where species are identified by frozensets with (key, value) pairs with the following keys.

    +----------------------+------------------------------------------------------------+
    | Species Type         | Identifier Keys                                            |
    +======================+============================================================+
    | ``molecule``         | ``name``, ``elements``, ``z``, ``electron_configuration``, |
    |                      | ``vibrational_level``, ``vibrational_mode``;               |
    +----------------------+------------------------------------------------------------+
    | ``molecular_bundle`` | ``name``, ``elements``, ``z_min``, ``z_max``;              |
    +----------------------+------------------------------------------------------------+
    | ``ion``              | ``name``, ``element``, ``z``, ``electron_configuration``;  |
    +----------------------+------------------------------------------------------------+
    | ``ion_bundle``       | ``name``, ``element``, ``z_min``, ``z_max``.               |
    +----------------------+------------------------------------------------------------+

    Parameters
    ----------
    profiles_struct : IDSStructure
        The IDS structure containing the core profiles data.

    Returns
    -------
    dict[str, dict[str, ndarray | None]]
        Dictionary with the species and their profiles.
    """

    species_types = ("molecule", "molecular_bundle", "ion", "ion_bundle")
    composition = {species_type: {} for species_type in species_types}

    composition["electron"] = load_core_profiles(profiles_struct.electrons)

    # ions
    ion_elements = []
    for ion in profiles_struct.ion:
        elements = tuple(get_element_list(ion.element))
        if not len(elements):
            raise RuntimeError("Unable to determine the ion species, ion.element AOS is empty.")
        ion_elements.append(elements)

        if len(ion.state):
            shared_temperature = _get_profile(ion, "temperature")
            backup_ids = None if len(ion.state) > 1 else ion
            for i, state in enumerate(ion.state):
                species_type, species_id = get_ion_state(state, i, elements)
                if species_id in composition[species_type]:
                    print(f"Warning! Skipping duplicated ion: {state.name.strip()}")
                    continue
                profiles = load_core_profiles(state, backup_ids)
                if backup_ids is None and profiles["temperature"] is None:
                    profiles["temperature"] = shared_temperature
                composition[species_type][species_id] = profiles
        else:
            species_type, species_id = get_ion(ion, elements)
            if species_id in composition[species_type]:
                print(f"Warning! Skipping duplicated ion: {ion.name.strip()}")
            else:
                composition[species_type][species_id] = load_core_profiles(ion)

    # neutrals
    for neutral in profiles_struct.neutral:
        elements = tuple(get_element_list(neutral.element))
        if not len(elements):
            elements = ion_elements[neutral.ion_index - 1]

        if len(neutral.state):
            shared_temperature = _get_profile(neutral, "temperature")
            backup_ids = None if len(neutral.state) > 1 else neutral
            for state in neutral.state:
                species_type, species_id = get_neutral_state(state, elements)
                if species_id in composition[species_type]:
                    print(f"Warning! Skipping duplicated neutral: {state.name.strip()}")
                    continue
                profiles = load_core_profiles(state, backup_ids)
                if backup_ids is None and profiles["temperature"] is None:
                    profiles["temperature"] = shared_temperature
                composition[species_type][species_id] = profiles
        else:
            species_type, species_id = get_neutral(neutral, elements)
            if species_id in composition[species_type]:
                print("Warning! Skipping duplicated neutral: ")
            else:
                composition[species_type][species_id] = load_core_profiles(neutral)

    # Replace missing species temperature with average ion temperature
    tion = profiles_struct.t_i_average
    if len(tion):
        for species_type in species_types:
            for species_id, profiles in composition[species_type].items():
                if profiles["temperature"] is None:
                    d = {first: second for first, second in species_id}
                    print(
                        "Warning! Using average ion temperature for the {} {}.".format(
                            d["name"], species_type
                        )
                    )
                    profiles["temperature"] = tion

    return composition


def _get_profile(ids_struct, name):
    if hasattr(ids_struct, name):
        arr = getattr(ids_struct, name)
        if len(arr):
            return np.asarray_chkfinite(arr)
